"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PlyFlow = exports.FlowResults = void 0;
const minimatch = require("minimatch");
const flowbee = require("flowbee");
const logger_1 = require("./logger");
const step_1 = require("./step");
const util = require("./util");
const names_1 = require("./names");
class FlowResults {
    constructor(name) {
        this.name = name;
        this.results = [];
    }
    get latest() {
        if (this.results.length > 0) {
            return this.results[this.results.length - 1];
        }
        else {
            return { name: this.name, status: 'Pending', message: '' };
        }
    }
    latestBad() {
        return this.latest.status === 'Failed' || this.latest.status === 'Errored';
    }
    get overall() {
        let otherThanSubmitted = false;
        for (const result of this.results) {
            if (result.status === 'Failed' || result.status === 'Errored') {
                return result;
            }
            else if (result.status !== 'Submitted') {
                otherThanSubmitted = true;
            }
        }
        return { name: this.name, status: otherThanSubmitted ? 'Passed' : 'Submitted', message: '' };
    }
}
exports.FlowResults = FlowResults;
class PlyFlow {
    constructor(flow, requestSuite, logger) {
        this.flow = flow;
        this.requestSuite = requestSuite;
        this.logger = logger;
        this.type = 'flow';
        this.start = 0;
        this._onFlow = new flowbee.TypedEvent();
        this.name = flowbee.getFlowName(flow);
        const id = util.genId();
        this.instance = {
            id,
            runId: id,
            flowPath: this.flow.path,
            status: 'Pending'
        };
        this.results = new FlowResults(this.name);
    }
    onFlow(listener) {
        this._onFlow.on(listener);
    }
    /**
     * Run a ply flow.
     */
    async run(runtime, values, runOptions) {
        var _a, _b, _c, _d;
        values[names_1.RUN_ID] = this.instance.runId || util.genId();
        if ((_a = this.flow.attributes) === null || _a === void 0 ? void 0 : _a.values) {
            const rows = JSON.parse((_b = this.flow.attributes) === null || _b === void 0 ? void 0 : _b.values);
            for (const row of rows) {
                values[row[0]] = row[1];
            }
        }
        // run values override even flow-configured vals
        if (runOptions === null || runOptions === void 0 ? void 0 : runOptions.values) {
            values = { ...values, ...runOptions.values };
        }
        if (((_c = this.flow.attributes) === null || _c === void 0 ? void 0 : _c.bail) === 'true') {
            runtime.options.bail = true;
        }
        runtime.results.actual.write('');
        const expectedExists = await this.requestSuite.runtime.results.expectedExists();
        if (!expectedExists && (runOptions === null || runOptions === void 0 ? void 0 : runOptions.submitIfExpectedMissing)) {
            runOptions.submit = true;
        }
        if ((runOptions === null || runOptions === void 0 ? void 0 : runOptions.createExpected) || (!expectedExists && (runOptions === null || runOptions === void 0 ? void 0 : runOptions.createExpectedIfMissing))) {
            this.logger.info(`Creating expected result: ${runtime.results.expected}`);
            runtime.results.expected.write('');
            runOptions.createExpected = true;
        }
        const runId = this.logger.level === logger_1.LogLevel.debug ? ` (${this.instance.runId})` : '';
        this.logger.info(`Running flow '${this.name}'${runId}`);
        this.instance.status = 'In Progress';
        this.instance.values = values;
        this.instance.start = new Date();
        this.emit('start', 'flow', this.instance);
        await this.runSubflows(this.getSubflows('Before'), runtime, values, runOptions);
        if (this.results.latestBad() && runtime.options.bail) {
            return this.endFlow();
        }
        const startStep = (_d = this.flow.steps) === null || _d === void 0 ? void 0 : _d.find(s => s.path === 'start');
        if (!startStep) {
            throw new Error(`Cannot find start step in flow: ${this.flow.path}`);
        }
        await this.exec(startStep, runtime, values, runOptions);
        if (this.results.latestBad() && runtime.options.bail) {
            return this.endFlow();
        }
        await this.runSubflows(this.getSubflows('After'), runtime, values, runOptions);
        return this.endFlow();
    }
    endFlow() {
        this.instance.end = new Date();
        if (this.instance.status === 'Errored' || this.instance.status === 'Failed') {
            this.emit('error', 'flow', this.instance);
        }
        else {
            this.instance.status = 'Completed';
            this.emit('finish', 'flow', this.instance);
        }
        return this.results.overall;
    }
    /**
     * Executes a step within a flow and recursively executes the following step(s).
     */
    async exec(step, runtime, values, runOptions, subflow) {
        var _a, _b, _c;
        await this.runSubflows(this.getSubflows('Before', step), runtime, values, runOptions);
        if (this.results.latestBad() && runtime.options.bail) {
            return;
        }
        const plyStep = new step_1.PlyStep(step, this.requestSuite, this.logger, this.flow.path, this.instance.id, subflow === null || subflow === void 0 ? void 0 : subflow.subflow);
        if (subflow) {
            if (!subflow.instance.stepInstances) {
                subflow.instance.stepInstances = [];
            }
            subflow.instance.stepInstances.push(plyStep.instance);
        }
        else {
            if (!this.instance.stepInstances) {
                this.instance.stepInstances = [];
            }
            this.instance.stepInstances.push(plyStep.instance);
        }
        this.emit('start', 'step', plyStep.instance);
        this.logger.info('Executing step', plyStep.name);
        this.emit('exec', 'step', plyStep.instance);
        this.results.results.push(await plyStep.run(runtime, values, runOptions));
        if (this.results.latestBad()) {
            this.instance.status = plyStep.instance.status;
            if (subflow)
                subflow.instance.status = plyStep.instance.status;
            this.emit('error', 'step', plyStep.instance);
            if (runtime.options.bail) {
                return;
            }
        }
        else {
            this.emit('finish', 'step', plyStep.instance);
        }
        await this.runSubflows(this.getSubflows('After', step), runtime, values, runOptions);
        if (this.results.latestBad() && runtime.options.bail) {
            return;
        }
        const outSteps = [];
        if (step.links) {
            for (const link of step.links) {
                const result = (_a = plyStep.instance.result) === null || _a === void 0 ? void 0 : _a.trim();
                if ((!result && !link.result) || (result === link.result)) {
                    let outStep;
                    if (subflow) {
                        outStep = (_b = subflow.subflow.steps) === null || _b === void 0 ? void 0 : _b.find(s => s.id === link.to);
                    }
                    else {
                        outStep = (_c = this.flow.steps) === null || _c === void 0 ? void 0 : _c.find(s => s.id === link.to);
                    }
                    if (!outStep) {
                        throw new Error(`No such step: ${link.to} (linked from ${link.id})`);
                    }
                    outSteps.push(outStep);
                }
            }
        }
        if (outSteps.length === 0 && step.path !== 'stop') {
            throw new Error(`No outbound link from step ${step.id} matches result: ${plyStep.instance.result}`);
        }
        // steps can execute in parallel
        await Promise.all(outSteps.map(outStep => this.exec(outStep, runtime, values, runOptions, subflow)));
    }
    getSubflows(type, step) {
        var _a;
        const subflows = ((_a = this.flow.subflows) === null || _a === void 0 ? void 0 : _a.filter(sub => {
            var _a, _b;
            if (((_a = sub.attributes) === null || _a === void 0 ? void 0 : _a.when) === type) {
                const steps = (_b = sub.attributes) === null || _b === void 0 ? void 0 : _b.steps;
                if (step) {
                    return steps ? minimatch(step.name, steps) : false;
                }
                else {
                    return !steps;
                }
            }
        })) || [];
        const flowInstanceId = this.instance.id;
        return subflows.map(subflow => {
            return {
                subflow,
                instance: {
                    id: Date.now().toString(16),
                    flowInstanceId,
                    subflowId: subflow.id,
                    runId: Date.now().toString(16),
                    flowPath: this.flow.path,
                    status: 'Pending'
                }
            };
        });
    }
    async runSubflows(subflows, runtime, values, runOptions) {
        var _a;
        for (const subflow of subflows) {
            const startStep = (_a = subflow.subflow.steps) === null || _a === void 0 ? void 0 : _a.find(s => s.path === 'start');
            if (!startStep) {
                throw new Error(`Cannot find start step in subflow: ${subflow.subflow.id}`);
            }
            subflow.instance.status = 'In Progress';
            subflow.instance.start = new Date();
            if (!this.instance.subflowInstances) {
                this.instance.subflowInstances = [];
            }
            this.instance.subflowInstances.push(subflow.instance);
            this.emit('start', 'subflow', subflow.instance);
            this.logger.info('Executing subflow', subflow.subflow.name);
            runtime.appendResult(`${subflow.subflow.name}:`, 0, runOptions === null || runOptions === void 0 ? void 0 : runOptions.createExpected, util.timestamp(subflow.instance.start));
            runtime.appendResult(`id: ${subflow.subflow.id}`, 1, runOptions === null || runOptions === void 0 ? void 0 : runOptions.createExpected);
            await this.exec(startStep, runtime, values, runOptions, subflow);
            subflow.instance.end = new Date();
            const elapsed = subflow.instance.end.getTime() - subflow.instance.start.getTime();
            if (this.results.latestBad()) {
                subflow.instance.status = this.results.latest.status === 'Errored' ? 'Errored' : 'Failed';
                runtime.appendResult(`status: ${subflow.instance.status}`, 1, runOptions === null || runOptions === void 0 ? void 0 : runOptions.createExpected, `${elapsed} ms`);
                this.emit('error', 'subflow', subflow.instance);
                if (runtime.options.bail) {
                    return;
                }
            }
            else {
                subflow.instance.status = 'Completed';
                runtime.appendResult(`status: ${subflow.instance.status}`, 1, runOptions === null || runOptions === void 0 ? void 0 : runOptions.createExpected, `${elapsed} ms`);
                this.emit('finish', 'subflow', subflow.instance);
            }
        }
    }
    flowEvent(eventType, elementType, instance) {
        return {
            eventType,
            elementType,
            flowPath: this.flow.path,
            flowInstanceId: this.instance.id,
            instance
        };
    }
    emit(eventType, elementType, instance) {
        this._onFlow.emit(this.flowEvent(eventType, elementType, instance));
    }
}
exports.PlyFlow = PlyFlow;
//# sourceMappingURL=flow.js.map