"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Location = void 0;
const fs = require("fs");
const path = require("path");
const util_1 = require("./util");
/**
 * Abstraction for a URL or file location.
 */
class Location {
    /**
     * @param path url or file path (backslashes are replaced with slashes)
     */
    constructor(path) {
        this.path = (0, util_1.fwdSlashes)(path);
        if (this.path.endsWith('/')) {
            this.path = this.path.substring(0, this.path.length - 1);
        }
    }
    get lastSlash() {
        const ls = this.path.lastIndexOf('/');
        if (ls !== -1) {
            return ls;
        }
    }
    /**
     * Undefined if no parent path
     */
    get parent() {
        if (this.lastSlash) {
            return this.path.substr(0, this.lastSlash);
        }
    }
    get name() {
        return this.lastSlash ? this.path.substring(this.lastSlash + 1) : this.path;
    }
    /**
     * Name without extension
     */
    get base() {
        const name = this.name;
        const lastDot = name.lastIndexOf('.');
        if (lastDot > 0 && lastDot < name.length - 1) {
            return name.substring(0, lastDot);
        }
        else {
            return name;
        }
    }
    get ext() {
        const name = this.name;
        const lastDot = name.lastIndexOf('.');
        if (lastDot > 0 && lastDot < name.length - 1) {
            return name.substring(lastDot + 1);
        }
    }
    /**
     * Returns zero for URLs, undefined if file does not exist
     */
    get timestamp() {
        if (this.isUrl) {
            return 0;
        }
        else if (fs.existsSync(this.path)) {
            return fs.statSync(this.path).mtimeMs;
        }
    }
    get isYaml() {
        return this.ext === 'yaml' || this.ext === 'yml';
    }
    get scheme() {
        if (this.path.startsWith('https://')) {
            return 'https';
        }
        else if (this.path.startsWith('http://')) {
            return 'http';
        }
    }
    get isUrl() {
        return !!this.scheme;
    }
    get isAbsolute() {
        return this.isUrl || path.isAbsolute(this.path);
    }
    get absolute() {
        if (this.isUrl) {
            return this.path;
        }
        else {
            return (0, util_1.fwdSlashes)(path.normalize(path.resolve(this.path)));
        }
    }
    /**
     * TODO: handle urls
     */
    isChildOf(parent) {
        const relative = this.relativeTo(parent);
        return relative.length > 0 && !relative.startsWith('..') && !path.isAbsolute(relative);
    }
    /**
     * TODO: handle urls
     */
    relativeTo(parent) {
        return (0, util_1.fwdSlashes)(path.normalize(path.relative(parent, this.path)));
    }
    toString() {
        return this.path;
    }
}
exports.Location = Location;
//# sourceMappingURL=location.js.map