"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RequestLoader = void 0;
const suite_1 = require("./suite");
const retrieval_1 = require("./retrieval");
const request_1 = require("./request");
const runtime_1 = require("./runtime");
const result_1 = require("./result");
const logger_1 = require("./logger");
const skip_1 = require("./skip");
const yaml = require("./yaml");
const util = require("./util");
class RequestLoader {
    constructor(locations, options) {
        this.locations = locations;
        this.options = options;
        if (options.skip) {
            this.skip = new skip_1.Skip(options.testsLocation, options.skip);
        }
    }
    async load() {
        const retrievals = this.locations.map(loc => new retrieval_1.Retrieval(loc));
        // load request files in parallel
        const promises = retrievals.map(retr => this.loadSuite(retr));
        const suites = await Promise.all(promises);
        suites.sort((s1, s2) => s1.name.localeCompare(s2.name));
        return suites;
    }
    sync() {
        const suites = [];
        for (const location of this.locations) {
            suites.push(this.syncSuite(new retrieval_1.Retrieval(location)));
        }
        suites.sort((s1, s2) => s1.name.localeCompare(s2.name));
        return suites;
    }
    async loadSuite(retrieval) {
        const contents = await retrieval.read();
        if (typeof contents === 'undefined') {
            throw new Error('Cannot retrieve: ' + retrieval.location.absolute);
        }
        const resultPaths = await result_1.ResultPaths.create(this.options, retrieval);
        return this.buildSuite(retrieval, contents, resultPaths);
    }
    syncSuite(retrieval) {
        const contents = retrieval.sync();
        if (typeof contents === 'undefined') {
            throw new Error('Cannot retrieve: ' + retrieval.location.absolute);
        }
        const resultPaths = result_1.ResultPaths.createSync(this.options, retrieval);
        return this.buildSuite(retrieval, contents, resultPaths);
    }
    buildSuite(retrieval, contents, resultPaths) {
        var _a;
        const runtime = new runtime_1.Runtime(this.options, retrieval, resultPaths);
        const logger = new logger_1.Logger({
            level: this.options.verbose ? logger_1.LogLevel.debug : (this.options.quiet ? logger_1.LogLevel.error : logger_1.LogLevel.info),
            prettyIndent: this.options.prettyIndent
        }, runtime.results.log);
        const suite = new suite_1.Suite(retrieval.location.base, 'request', retrieval.location.relativeTo(this.options.testsLocation), runtime, logger, 0, util.lines(contents).length - 1);
        const obj = yaml.load(retrieval.location.path, contents, true);
        if (obj) {
            for (const key of Object.keys(obj)) {
                const val = obj[key];
                if (typeof val === 'object') {
                    const startEnd = { start: val.__start, end: val.__end };
                    const { __start, __end, ...cleanObj } = val;
                    const request = new request_1.PlyRequest(key, { ...startEnd, ...cleanObj }, logger, retrieval);
                    suite.add(request);
                }
            }
        }
        // mark if skipped
        if ((_a = this.skip) === null || _a === void 0 ? void 0 : _a.isSkipped(suite.path)) {
            suite.skip = true;
        }
        return suite;
    }
}
exports.RequestLoader = RequestLoader;
//# sourceMappingURL=requests.js.map