"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DecoratedSuite = exports.Runtime = void 0;
const path = require("path");
const minimatch = require("minimatch");
const location_1 = require("./location");
const names_1 = require("./names");
/**
 * Runtime information for a test suite.
 */
class Runtime {
    constructor(options, retrieval, results) {
        this.options = options;
        this.retrieval = retrieval;
        this.results = results;
        if (path.isAbsolute(this.options.testsLocation)) {
            this.testsLocation = new location_1.Location(this.options.testsLocation);
        }
        else {
            this.testsLocation = new location_1.Location(path.resolve(process.cwd() + '/' + this.options.testsLocation));
        }
    }
    get suitePath() {
        return this.retrieval.location.relativeTo(this.options.testsLocation);
    }
    appendResult(line, level = 0, withExpected = false, comment) {
        line = line.padStart(line.length + level * (this.options.prettyIndent || 0));
        this.results.actual.append(`${line}${comment ? '  # ' + comment : ''}\n`);
        if (withExpected) {
            this.results.expected.append(`${line}\n`);
        }
    }
}
exports.Runtime = Runtime;
/**
 * Applicable for Cases
 */
class DecoratedSuite {
    /**
     * @param instance runtime instance of a suite
     */
    constructor(instance) {
        this.instance = instance;
        this.testCases = [];
        this.befores = [];
        this.afters = [];
        this.testSuite = instance.constructor[names_1.SUITE];
        if (this.testSuite) {
            this.testSuite = { ...this.testSuite, className: this.testSuite.name };
        }
        Object.getOwnPropertyNames(instance.constructor.prototype).forEach(propName => {
            try {
                if (typeof instance.constructor.prototype[propName] === 'function') {
                    const method = instance.constructor.prototype[propName];
                    if (method[names_1.TEST]) {
                        const testCase = method[names_1.TEST];
                        if (!this.testCases.find(tc => tc.name === testCase.name)) {
                            this.testCases.push({ ...testCase, method });
                        }
                    }
                    if (method[names_1.BEFORE]) {
                        const before = method[names_1.BEFORE];
                        if (!this.befores.find(b => b.name === before.name)) {
                            this.befores.push({ ...before, method });
                        }
                    }
                    if (method[names_1.AFTER]) {
                        const after = method[names_1.AFTER];
                        if (!this.afters.find(a => a.name === after.name)) {
                            this.afters.push({ ...after, method });
                        }
                    }
                }
            }
            catch (_ignored) {
                // getter or setter before constructor?
            }
        });
    }
    async runIfMatch(beforeOrAfter, test, values) {
        if (this.isMatch(beforeOrAfter, test)) {
            if (beforeOrAfter.tests || !beforeOrAfter.hasRun) {
                await beforeOrAfter.method.call(this.instance, values);
                beforeOrAfter.hasRun = true;
            }
        }
    }
    isMatch(beforeOrAfter, test) {
        return !beforeOrAfter.tests || minimatch(test, beforeOrAfter.tests);
    }
    async runBefores(test, values) {
        for (const before of this.befores) {
            await this.runIfMatch(before, test, values);
        }
    }
    async runAfters(test, values) {
        for (const after of this.afters) {
            await this.runIfMatch(after, test, values);
        }
    }
}
exports.DecoratedSuite = DecoratedSuite;
//# sourceMappingURL=runtime.js.map