"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Storage = void 0;
const fs = require("fs");
const os = require("os");
const mkdirp = require("mkdirp");
const util = require("./util");
const location_1 = require("./location");
/**
 * Abstracts storage to file system or html localStorage.
 */
class Storage {
    /**
     * @param location file or logical path
     */
    constructor(location) {
        this.location = new location_1.Location(location);
        if (typeof localStorage !== 'undefined') {
            this.localStorage = localStorage;
        }
    }
    get exists() {
        if (this.localStorage) {
            return this.localStorage.getItem(this.location.path) !== null;
        }
        else {
            return fs.existsSync(this.location.path);
        }
    }
    read() {
        if (this.localStorage) {
            return this.localStorage.getItem(this.location.path);
        }
        else {
            if (fs.existsSync(this.location.path)) {
                return fs.readFileSync(this.location.path, 'utf-8');
            }
        }
    }
    /**
     * For file system storage, all newlines are replaced with OS-appropriate
     */
    write(contents) {
        if (this.localStorage) {
            this.localStorage.setItem(this.location.path, contents);
        }
        else {
            if (this.location.parent) {
                mkdirp.sync(this.location.parent);
            }
            fs.writeFileSync(this.location.path, contents.replace(/\r?\n/, os.EOL));
        }
    }
    /**
     * For file system storage, all newlines are replaced with OS-appropriate
     */
    append(contents) {
        if (this.localStorage) {
            const exist = this.localStorage.getItem(this.location.path);
            this.localStorage.setItem(this.location.path, exist ? exist + contents : contents);
        }
        else {
            if (this.location.parent) {
                mkdirp.sync(this.location.parent);
            }
            fs.appendFileSync(this.location.path, contents.replace(/\r?\n/, os.EOL));
        }
    }
    insert(contents, start) {
        let newLines = util.lines(contents);
        const exist = this.read();
        if (exist) {
            const existLines = util.lines(exist);
            const preLines = start > 0 ? existLines.slice(0, start) : [];
            const postLines = existLines.slice(start);
            newLines = [...preLines, ...newLines, ...postLines];
        }
        this.write(newLines.join('\n'));
    }
    padLines(start, lines) {
        this.insert(''.padStart(lines - 1, '\n'), start);
    }
    clear() {
        this.write('');
    }
    remove() {
        if (this.localStorage) {
            this.localStorage.removeItem(this.location.path);
        }
        else {
            if (fs.existsSync(this.location.path)) {
                fs.unlinkSync(this.location.path);
            }
        }
    }
    toString() {
        return this.location.toString();
    }
}
exports.Storage = Storage;
//# sourceMappingURL=storage.js.map