"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DefaultRowConverter = exports.fromXlsx = exports.fromCsv = exports.Values = void 0;
const fs = require("fs");
const path = require("path");
const stream = require("stream");
const deepmerge = require("deepmerge");
const csv = require("csv-parse");
const transform = require("stream-transform");
const node_1 = require("read-excel-file/node");
const retrieval_1 = require("./retrieval");
/**
 * Environment variable for additional values
 */
const PLY_VALUES = 'PLY_VALUES';
class Values {
    constructor(locations, logger) {
        this.locations = locations;
        this.logger = logger;
        for (const loc of locations) {
            if (loc.endsWith('.csv') || loc.endsWith('.xlsx')) {
                if (this.rowsLoc) {
                    throw new Error('Only one values file may be .csv or .xlsx');
                }
                this.rowsLoc = loc;
            }
        }
    }
    get isRows() {
        return !!this.rowsLoc;
    }
    async read() {
        let values = {};
        for (const location of this.locations) {
            if (location.endsWith('.csv') || location.endsWith('.xlsx')) {
                this.logger.debug(`Delayed load rowwise values file" ${location}`);
            }
            else {
                const contents = await new retrieval_1.Retrieval(location).read();
                if (contents) {
                    try {
                        const obj = JSON.parse(contents);
                        values = deepmerge(values, obj);
                    }
                    catch (err) {
                        throw new Error(`Cannot parse values file: ${location} (${err.message})`);
                    }
                }
                else {
                    this.logger.debug(`Values file not found: ${path.normalize(path.resolve(location))}`);
                }
            }
        }
        this.logger.debug('Values (excluding PLY_VALUES env var)', values);
        const envValues = process.env[PLY_VALUES];
        if (envValues) {
            try {
                const obj = JSON.parse(envValues);
                values = deepmerge(values, obj);
            }
            catch (err) {
                throw new Error(`Cannot parse ${PLY_VALUES} (${err.message})`);
            }
        }
        return values;
    }
    async getRowStream() {
        if (!this.rowsLoc) {
            throw new Error('Rowwise values required for row converter');
        }
        const baseVals = await this.read() || {};
        if (this.rowsLoc.endsWith('.xlsx')) {
            const readable = new stream.Readable({ objectMode: true });
            for (const row of await (0, exports.fromXlsx)(this.rowsLoc)) {
                this.logger.debug('Row values', row);
                readable.push(deepmerge(baseVals, row));
            }
            readable.push(null);
            return readable;
        }
        else {
            // stream csv records
            const parser = fs.createReadStream(this.rowsLoc)
                .pipe(csv({ to_line: 1 }));
            // header row
            let converter;
            for await (const row of parser) {
                converter = new DefaultRowConverter(row);
            }
            const transformer = transform(async (row, cb) => {
                const converted = converter.convert(row);
                this.logger.debug('Row values', converted);
                cb(null, deepmerge(baseVals, converted));
            });
            return fs
                .createReadStream(this.rowsLoc)
                .pipe(csv({ from_line: 2 }))
                .pipe(transformer);
        }
    }
}
exports.Values = Values;
/**
 * Reads entire csv file into rows in memory
 */
const fromCsv = async (file) => {
    const valueObjs = [];
    const parser = fs
        .createReadStream(file)
        .pipe(csv({
    // CSV options if any
    }));
    let converter;
    for await (const row of parser) {
        if (converter) {
            valueObjs.push(converter.convert(row));
        }
        else {
            converter = new DefaultRowConverter(row);
        }
    }
    return valueObjs;
};
exports.fromCsv = fromCsv;
/**
 * Reads entire xlsx file into rows in memory
 */
const fromXlsx = async (file) => {
    const valueObjs = [];
    const rows = await (0, node_1.default)(file);
    let converter;
    for await (const row of rows) {
        if (converter) {
            valueObjs.push(converter.convert(row));
        }
        else {
            converter = new DefaultRowConverter(row, { inferPrimitiveTypes: false });
        }
    }
    return valueObjs;
};
exports.fromXlsx = fromXlsx;
const defaultOptions = {
    trimValues: true,
    trimLabels: true,
    inferPrimitiveTypes: true,
    blankIsNull: true
};
class DefaultRowConverter {
    constructor(names, options) {
        this.options = deepmerge(defaultOptions, options || {});
        this.names = names.map(name => {
            if (this.options.trimLabels) {
                return ('' + name).trim();
            }
            else {
                return '' + name;
            }
        });
    }
    splitName(name) {
        return name.split('.');
    }
    convert(row) {
        const obj = {};
        for (let i = 0; i < row.length; i++) {
            const segs = this.splitName(this.names[i]);
            let cur = obj;
            for (let j = 0; j < segs.length; j++) {
                const seg = segs[j];
                let key = seg;
                let arrIdx;
                if (seg.endsWith(']')) {
                    const sq1 = seg.indexOf('[');
                    if (sq1 !== -1) {
                        key = seg.substring(0, sq1);
                        arrIdx = parseInt(seg.substring(sq1 + 1, seg.length - 1));
                    }
                }
                if (j === segs.length - 1) {
                    if (typeof arrIdx === 'number') {
                        if (typeof cur[key] === 'undefined') {
                            cur[key] = [];
                        }
                        cur[key][arrIdx] = this.getValue(row[i]);
                    }
                    else {
                        cur[key] = this.getValue(row[i]);
                    }
                }
                else if (typeof arrIdx === 'number') {
                    if (typeof cur[key] === 'undefined') {
                        cur[key] = [];
                    }
                    if (typeof cur[key][arrIdx] === 'undefined') {
                        cur[key][arrIdx] = {};
                    }
                    cur = cur[key][arrIdx];
                }
                else {
                    if (typeof cur[key] === 'undefined') {
                        cur[key] = {};
                    }
                    cur = cur[key];
                }
            }
        }
        return obj;
    }
    /**
     * TODO: Date
     */
    getValue(str) {
        let val = str;
        if (typeof val === 'string') {
            if (this.options.trimValues) {
                val = val.trim();
            }
            if (this.options.inferPrimitiveTypes) {
                const num = parseFloat(val);
                if (!isNaN(num))
                    return num;
                if (val.toLowerCase() === 'true')
                    return true;
                else if (val.toLowerCase() === 'false')
                    return false;
            }
            if (this.options.blankIsNull && val === '')
                return null;
            if (val === "''")
                return '';
        }
        return val;
    }
}
exports.DefaultRowConverter = DefaultRowConverter;
//# sourceMappingURL=values.js.map